/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)
https://sites.google.com/site/treesplicer/

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.

These benchmarks are adapted from the Lonestar benchmark suite:
http://iss.ices.utexas.edu/?p=projects/galois/lonestar
 */

package barneshut;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.Locale;
import java.util.Random;
import java.util.Scanner;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicLong;
import java.util.zip.GZIPInputStream;

import util.Launcher;

/**
 * Main class of the Barnes Hut application.
 */
public final class Main {
	private static int nbodies; // number of bodies in system
	private static int ntimesteps; // number of time steps to run
	private static double dtime; // length of one time step
	private static double eps; // potential softening parameter
	private static double tol; // tolerance for stopping recursion, should be less than 0.57 for 3D case to bound error

	private static double dthf, epssq, itolsq;
	private static OctTreeLeafNodeData body[]; // the n bodies
	private static OctTreeLeafNodeData[] sorted_body;
	private static double diameter, centerx, centery, centerz;
	private static int curr;

	static boolean sort;

	/*
	 * Reads the input from a gzipped or clear text file.
	 */
	private static void ReadInput(String filename, boolean print) {
		double vx, vy, vz;
		Scanner in = null;
		try {
			if (filename.endsWith(".gz")) {
				in = new Scanner(new GZIPInputStream(new FileInputStream(filename)));
			} else {
				in = new Scanner(new BufferedReader(new FileReader(filename)));
			}
			in.useLocale(Locale.US);
		} catch (FileNotFoundException e) {
			System.err.println(e);
			System.exit(-1);
		} catch (IOException e) {
			System.err.println(e);
			System.exit(-1);
		}
		nbodies = in.nextInt();
		ntimesteps = in.nextInt();
		dtime = in.nextDouble();
		eps = in.nextDouble();
		tol = in.nextDouble();
		dthf = 0.5 * dtime;
		epssq = eps * eps;
		itolsq = 1.0 / (tol * tol);
		if (print) {
			System.err.println("configuration: " + nbodies + " bodies, " + ntimesteps + " time steps");
			System.err.println();
		}
		body = new OctTreeLeafNodeData[nbodies];
		for (int i = 0; i < nbodies; i++) {
			body[i] = new OctTreeLeafNodeData();
		}
		for (int i = 0; i < nbodies; i++) {
				body[i].mass = in.nextDouble();
				body[i].posx = in.nextDouble();
				body[i].posy = in.nextDouble();
				body[i].posz = in.nextDouble();
				vx = in.nextDouble();
				vy = in.nextDouble();
				vz = in.nextDouble();
				body[i].setVelocity(vx, vy, vz);
		}
	}
	
	private static void RandomInput(int num, boolean print) {
		double vx, vy, vz;
		Random rand = new Random(0);
		nbodies = num;
		ntimesteps = 1;
		dtime = 0.025;
		eps = 0.05;
		tol = 0.5;
		dthf = 0.5 * dtime;
		epssq = eps * eps;
		itolsq = 1.0 / (tol * tol);
		if (print) {
			System.err.println("configuration: " + nbodies + " bodies, " + ntimesteps + " time steps");
			System.err.println();
		}
		body = new OctTreeLeafNodeData[nbodies];
		for (int i = 0; i < nbodies; i++) {
			body[i] = new OctTreeLeafNodeData();
		}
		for (int i = 0; i < nbodies; i++) {
				body[i].mass = 1.0E-04;
				body[i].posx = rand.nextDouble();
				body[i].posy = rand.nextDouble();
				body[i].posz = rand.nextDouble();
				vx = rand.nextDouble();
				vy = rand.nextDouble();
				vz = rand.nextDouble();
				body[i].setVelocity(vx, vy, vz);
		}
	}

	/*
	 * Computes a bounding box around all the bodies.
	 */
	private static void ComputeCenterAndDiameter() {
		double minx, miny, minz;
		double maxx, maxy, maxz;
		double posx, posy, posz;
		minx = miny = minz = Double.MAX_VALUE;
		maxx = maxy = maxz = Double.MIN_VALUE;
		for (int i = 0; i < nbodies; i++) {
			posx = body[i].posx;
			posy = body[i].posy;
			posz = body[i].posz;
			if (minx > posx) {
				minx = posx;
			}
			if (miny > posy) {
				miny = posy;
			}
			if (minz > posz) {
				minz = posz;
			}
			if (maxx < posx) {
				maxx = posx;
			}
			if (maxy < posy) {
				maxy = posy;
			}
			if (maxz < posz) {
				maxz = posz;
			}
		}
		diameter = maxx - minx;
		if (diameter < (maxy - miny)) {
			diameter = (maxy - miny);
		}
		if (diameter < (maxz - minz)) {
			diameter = (maxz - minz);
		}
		centerx = (maxx + minx) * 0.5;
		centery = (maxy + miny) * 0.5;
		centerz = (maxz + minz) * 0.5;
	}

	/*
	 * Recursively inserts a body into the octree.
	 */
	private static void Insert(Node root, OctTreeLeafNodeData b,
			double r) {
		double x = 0.0, y = 0.0, z = 0.0;
		OctTreeNodeData n = root.getData();
		int i = 0;
		if (n.posx < b.posx) {
			i = 1;
			x = r;
		}
		if (n.posy < b.posy) {
			i += 2;
			y = r;
		}
		if (n.posz < b.posz) {
			i += 4;
			z = r;
		}
		Node child = root.getChild(i);
		if (child == null) {
			Node newnode = new Node(b);
			root.setChild(i, newnode);
		} else {
			double rh = 0.5 * r;
			OctTreeNodeData ch = child.getData();
			if (!(ch.isLeaf())) {
				Insert(child, b, rh);
			} else {
				Node newnode = new Node(new OctTreeNodeData(n.posx - rh + x, n.posy - rh + y, n.posz
						- rh + z));
				Insert(newnode, b, rh);
				Insert(newnode, (OctTreeLeafNodeData) ch, rh);
				root.setChild(i, newnode);
			}
		}
	}

	/*
	 * Traverses the tree bottom up to compute the total mass and the center of
	 * mass of all the bodies in the subtree rooted in each internal octree node.
	 */
	private static void ComputeCenterOfMass(Node root) {
		double m, px = 0.0, py = 0.0, pz = 0.0;
		OctTreeNodeData n = root.getData();
		int j = 0;
		n.mass = 0.0;
		for (int i = 0; i < 8; i++) {
			Node child = root.getChild(i);
			if (child != null) {
				if (i != j) {
					root.setChild(i, null);
					root.setChild(j, child);
				}
				j++;
				OctTreeNodeData ch = child.getData();
				if (ch.isLeaf()) {
					if (sort) {
						sorted_body[curr++] = (OctTreeLeafNodeData) ch;
					}
				} else {
					ComputeCenterOfMass(child);
				}
				m = ch.mass;
				n.mass += m;
				px += ch.posx * m;
				py += ch.posy * m;
				pz += ch.posz * m;
			}
		}
		m = 1.0 / n.mass;
		n.posx = px * m;
		n.posy = py * m;
		n.posz = pz * m;
	}

	/*
	 * Calculates the force acting on a body
	 */
	private static void ComputeForce(OctTreeLeafNodeData nd, Node root, double size, double itolsq, int step, double dthf, double epssq) {
		double ax, ay, az;
		ax = nd.accx;
		ay = nd.accy;
		az = nd.accz;
		nd.accx = 0.0;
		nd.accy = 0.0;
		nd.accz = 0.0;
		RecurseForce(nd, root, size * size * itolsq, epssq);
		if (step > 0) {
			nd.velx += (nd.accx - ax) * dthf;
			nd.vely += (nd.accy - ay) * dthf;
			nd.velz += (nd.accz - az) * dthf;
		}
	}

	/*
	 * Helper method to compute the force acting on a body (recursively walks the
	 * tree)
	 */
	public static void RecurseForce(OctTreeLeafNodeData nd, Node nn, double dsq, double epssq) {
		if (nn == null) {
			return;
		}
		double drx, dry, drz, drsq, nphi, scale, idr;
		OctTreeNodeData n = nn.getData();
		drx = n.posx - nd.posx;
		dry = n.posy - nd.posy;
		drz = n.posz - nd.posz;
		drsq = drx * drx + dry * dry + drz * drz;
		nd.nodesTraversed++;
		if (drsq < dsq) {
			if (!(n.isLeaf())) { // n is a cell
				RecurseForce(nd, nn.child0, dsq * 0.25, epssq);
				RecurseForce(nd, nn.child1, dsq * 0.25, epssq);
				RecurseForce(nd, nn.child2, dsq * 0.25, epssq);
				RecurseForce(nd, nn.child3, dsq * 0.25, epssq);
				RecurseForce(nd, nn.child4, dsq * 0.25, epssq);
				RecurseForce(nd, nn.child5, dsq * 0.25, epssq);
				RecurseForce(nd, nn.child6, dsq * 0.25, epssq);
				RecurseForce(nd, nn.child7, dsq * 0.25, epssq);
			} else { // n is a body
				if (n != nd) {
					drsq += epssq;
					idr = 1 / Math.sqrt(drsq);
					nphi = n.mass * idr;
					scale = nphi * idr * idr;
					nd.accx += drx * scale;
					nd.accy += dry * scale;
					nd.accz += drz * scale;
				}
			}
		} else { // node is far enough away, don't recurse any deeper
			drsq += epssq;
			idr = 1 / Math.sqrt(drsq);
			nphi = n.mass * idr;
			scale = nphi * idr * idr;
			nd.accx += drx * scale;
			nd.accy += dry * scale;
			nd.accz += drz * scale;
		}
	}

	/*
	 * Advances a body's velocity and position by one time step
	 */
	private static void Advance(double dthf, double dtime) {
		double dvelx, dvely, dvelz;
		double velhx, velhy, velhz;

		for (int i = 0; i < nbodies; i++) {
			OctTreeLeafNodeData nd = body[i];
			body[i] = nd;
			dvelx = nd.accx * dthf;
			dvely = nd.accy * dthf;
			dvelz = nd.accz * dthf;
			velhx = nd.velx + dvelx;
			velhy = nd.vely + dvely;
			velhz = nd.velz + dvelz;
			nd.posx += velhx * dtime;
			nd.posy += velhy * dtime;
			nd.posz += velhz * dtime;
			nd.velx = velhx + dvelx;
			nd.vely = velhy + dvely;
			nd.velz = velhz + dvelz;
		}
	}

	/**
	 * Main method to launch the binary search tree code.
	 */
	public static void main(String args[]) throws ExecutionException {
		
		if (args.length != 3) {
			System.err.println("arguments: sort randomize [inputFile / numBodies]");
			System.err.println("ex) true false BarnesHutA.in.gz");
			System.err.println("ex) true true 100000");
			System.exit(-1);
		}
		sort = Boolean.parseBoolean(args[0]);
		boolean randomize = Boolean.parseBoolean(args[1]);
		if (randomize) {
			RandomInput(Integer.parseInt(args[2]), true);
		} else {
			ReadInput(args[2], true);	
		}
		
		OctTreeNodeData res = null;

		for (int step = 0; step < ntimesteps; step++) { // time-step the system
			final int s = step;
			ComputeCenterAndDiameter();
			final Node root = new Node(new OctTreeNodeData(centerx, centery, centerz)); // create the
			double radius = diameter * 0.5;
			for (int i = 0; i < nbodies; i++) {
				Insert(root, body[i], radius); // grow the tree by inserting each body
			}
			curr = 0;
			if (sort) {
				sorted_body = new OctTreeLeafNodeData[nbodies];
			}   
			// summarize subtree info in each internal node (plus restructure tree and sort bodies for performance reasons)
			ComputeCenterOfMass(root);
			if (sort) {
				body = sorted_body;
			}

			Launcher.startTiming();
			ParallelFor(root, step);
			Launcher.stopTiming();
			
			Advance(dthf, dtime); // advance the position and velocity of each body

			// print center of mass for this timestep
			res = root.getData();
			System.out.println("Timestep " + step + " Center of Mass = " + res.posx + " " + res.posy + " " + res.posz);

			long totalNodesTraversed = 0;
			for (int i = 0; i < body.length; i++) {
				totalNodesTraversed += body[i].nodesTraversed;
			}
			System.out.printf("totalNodesTraversed: %d\n", totalNodesTraversed);

			if (nbodies == 10000 && randomize) {
				// check expected values				
				if (s == 0) {
					assert totalNodesTraversed == 4586829L;
					assert res.posx == 0.4995124926425938D;
					assert res.posy == 0.4971952727192604D;
					assert res.posz == 0.5038525463784792D;
				}
				if (s == 1) {
					assert totalNodesTraversed == 9345701L;
					assert res.posx == 0.5120021242194736D;
					assert res.posy == 0.5097297518450723D;
					assert res.posz == 0.5163728449901297D;
				}
				System.out.println("expected values passed");
			}
		} 
	}
	
	// a substitute for parallel for annotations on for statements, which are not supported in Java 6
	public @interface ParallelFor {}
	@ParallelFor
	public static void ParallelFor(Node root, int s) {
		for (int i = 0; i < body.length; i++) {
			OctTreeLeafNodeData item = body[i];
			ComputeForce(item, root, diameter, itolsq, s, dthf, epssq);
		}
	}
}
