/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)
https://sites.google.com/site/treesplicer/

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.
 */

package knearestneighborballtreeopt;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Locale;
import java.util.Random;
import java.util.Scanner;
import java.util.zip.GZIPInputStream;

import util.Launcher;

public class Main {

	static final boolean verify = false;
	static int DIM, K, maxLabels;

	public static void main(String [] args) throws Exception {
		if (args.length != 7) {
			System.err.println("arguments: sort dimensionality K numLabels numTrainPoints numTestPoints [filename]");
			System.err.println("ex) false 3 5 1 10000 10000 random");
			System.err.println("ex) false 7 5 8 10000 10000 input/nearestneighbor/covtype.7d.gz");
			System.exit(-1);
		}
		boolean sort = Boolean.parseBoolean(args[0]);
		DIM = Integer.parseInt(args[1]);
		K = Integer.parseInt(args[2]);
		maxLabels = Integer.parseInt(args[3]);
		int numTrainPoints = Integer.parseInt(args[4]);
		int numTestPoints = Integer.parseInt(args[5]);
		String filename = args[6];
		
		Attribute [] attributeArray = new Attribute[DIM];
		ArrayList<Attribute> attributes = new ArrayList<Attribute>(DIM);
		for (int i = 0; i < DIM; i++) {
			attributeArray[i] = new Attribute("val" + i); 		
			attributes.add(attributeArray[i]);
		}
		Instances insts = new Instances("points", attributes, numTrainPoints);
		Instance [] trainData = new Instance[numTrainPoints];
		Instance [] testData = new Instance[numTestPoints];

		if (filename.equals("random")) {
			Random rand = new Random(0);
			genData(attributeArray, trainData, rand);
			genData(attributeArray, testData, rand);			
		} else {
			readInput(attributeArray, filename, trainData, testData);
		}
		float rate = process(insts, trainData, testData, sort, filename);
		recordCorrectRate(args, rate);
	}

	static void recordCorrectRate(String [] args, float rate) {
		String argstr = "";
		for (int i = 0; i < args.length; i++) {
			argstr += args[i] + " ";
		} 
		PrintStream csv;
		try {
			csv = new PrintStream(new FileOutputStream("nncorrect.csv", true));
			csv.printf("%s, %.4f\n", argstr, rate);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	static float process(Instances insts, final Instance[] trainData, final Instance[] testData, boolean sort, String filename) throws Exception {

		if (sort) {
			sort(testData, 0, testData.length);
		}

		NormalizableDistance distanceFunction = new EuclideanDistance();
		distanceFunction.setDontNormalize(true);
		distanceFunction.setInstances(insts);
		double [] naiveClosests = null;
		int [] labelResults = new int[testData.length];
		Instance [] closestResults = new Instance[testData.length];

		for (int i = 0; i < trainData.length; i++) {
			insts.add(trainData[i]);
		}

		BallTree ballTree = new BallTree();
		ballTree.setDistanceFunction(distanceFunction);
		ballTree.setInstances(insts);
		Launcher.startTiming();
		ParallelFor(ballTree, testData, closestResults, labelResults);
		Launcher.stopTiming();

		int correct = 0;
		int numNodesTraversed = 0;
		for (int i = 0 ; i < testData.length; i++) {
			DenseInstance di0 = (DenseInstance) testData[i];
			if (di0.label == labelResults[i]) correct++;
			numNodesTraversed += di0.nodesTraversed;
		}
		float correctRate = ((float) correct) / testData.length;
		System.out.printf("correct: %f\n", correctRate);
		System.out.printf("nodesTraversed: %d\n", numNodesTraversed);
		
		if (trainData.length == 10000 && testData.length == 10000 && K == 1 && DIM == 7 && filename.equals("random")) {
			if (numNodesTraversed == 3790278) {
				System.out.println("expected values passed");
			} else {
				System.out.println("expected values failed");
				assert false;
			}
		}
		
		if (verify) {
			naiveClosests = new double[testData.length];
			for (int i = 0; i < testData.length; i++) {
				naiveClosests[i] = naiveNearest(testData[i], trainData, distanceFunction);
			}
			for (int i = 0 ; i < testData.length; i++) {
				double closestDist = distanceFunction.distance(testData[i], closestResults[i]);
				if (closestDist != naiveClosests[i]) {
					System.out.printf("%f %f\n", closestDist, naiveClosests[i]);
					assert false;
				}
			}
		}
		return correctRate;
	}

	// a substitute for parallel for annotations on for statements, which are not supported in Java 6
	public @interface ParallelFor {}
	@ParallelFor
	public static void ParallelFor(final BallTree ballTree, final Instance [] testData, 
			final Instance [] closestResults, final int [] labelResults) throws Exception {
		
		for (int i = 0; i < testData.length; i++) {
			Instance query = testData[i];
			Instances closests = ballTree.kNearestNeighbours(query, K);
			closestResults[i] = closests.instance(0);
			labelResults[i] = majorityLabel(closests);
		}
	}
	
	public static int majorityLabel(Instances insts) {
		int [] cnts = new int[maxLabels];
		for (int i = 0; i < K; i++) {
			DenseInstance di = (DenseInstance) insts.instance(i); 
			int label =  di.label;
			cnts[label]++;
		}
		int label = 0;
		for (int i = 0; i < maxLabels; i++) {
			if (cnts[i] > cnts[label]) {
				label = i;
			}
		}
		return label;
	}
	
	public static class InstanceComparable implements Comparator<Instance> {
		int type;
		
		public InstanceComparable(int type) {
			this.type = type;
		}

		@Override
		public int compare(Instance p0, Instance p1) {
			double val0 = p0.value(type);
			double val1 = p1.value(type);
			return val0 < val1 ? -1 : (val0 == val1 ? 0 : 1); 
		}	
	}
	
	public static void sort(Instance [] list, int offset, int size) {
		if (size < 5) return;
		//compute bounding box of points
		double [] min = new double[DIM];
		double [] max = new double[DIM];
		for (int i = 0; i < DIM; i++) {
			min[i] = Double.MAX_VALUE;
			max[i] = Double.MIN_VALUE;
		}
		for (int i = offset; i < size + offset; i++) {    	
			for (int j = 0; j < DIM; j++) {
				Instance inst = list[i];
				assert inst != null;
				min[j] = Math.min(inst.value(j), min[j]);
				max[j] = Math.max(inst.value(j), max[j]);
			}
		}
		
		//choose split plane
		double [] diff = new double[DIM];
		for (int i = 0; i < DIM; i++) {
			diff[i] = max[i] - min[i];
		}
		int type = 0;
		double maxDiff = 0;
		for (int i = 0; i < DIM; i++) {
			if (diff[i] > maxDiff) {
				maxDiff = diff[i];
				type = i;
			}
		}
		
		Arrays.sort(list, offset, offset + size, new InstanceComparable(type));
		int leftCount = size / 2;
		sort(list, offset, leftCount);
		sort(list, offset + leftCount, size - leftCount);
	}

	public static double naiveNearest(Instance query, Instance [] trainData, DistanceFunction distanceFunction) {
		Instance closest = null;
		double closestDist = 0;

		for (int i = 0; i < trainData.length; i++) {
			double dist = distanceFunction.distance(query, trainData[i]);
			if (closest == null || dist < closestDist) {
				closest = trainData[i];
				closestDist = dist;
			}
		}
		return closestDist;
	}

	static void readInput(Attribute [] attributeArray, String filename, Instance [] trainData, Instance [] testData) {
		Scanner in = null;
		try {
			if (filename.endsWith(".gz")) {
				in = new Scanner(new GZIPInputStream(new FileInputStream(filename)));
			} else {
				in = new Scanner(new BufferedReader(new FileReader(filename)));
			}
			in.useLocale(Locale.US);
		} catch (FileNotFoundException e) {
			System.err.println(e);
			System.exit(-1);
		} catch (IOException e) {
			System.err.println(e);
			System.exit(-1);
		}
		readData(attributeArray, trainData, in);
		readData(attributeArray, testData, in);
	}

	static void readData(Attribute [] attributeArray, Instance [] data, Scanner in) {
		int idx = 0;
		do {
			DenseInstance inst = new DenseInstance(DIM);
			inst.label = in.nextInt();
			assert inst.label < maxLabels;
			for (int i = 0; i < DIM; i++) {
				inst.setValue(attributeArray[i], in.nextDouble());
			}
			data[idx++] = inst;
		} while (idx < data.length);	
	}
	
	public static void genData(Attribute [] attributeArray, Instance [] insts, Random rand) {
		for (int i = 0; i < insts.length; i++) {
			Instance inst = new DenseInstance(DIM);   
			for (int j = 0; j < DIM; j++) {
				inst.setValue(attributeArray[j], rand.nextDouble());  
			}
			insts[i] = inst;
		}
	}
}
