/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)
https://sites.google.com/site/treesplicer/

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.

These benchmarks are adapted from the Lonestar benchmark suite:
http://iss.ices.utexas.edu/?p=projects/galois/lonestar
 */

package knearestneighborsplitplane;

import java.util.Arrays;
import java.util.Collections;
import java.util.concurrent.atomic.AtomicLong;

public class KdCell {

	public static final int DIM = Main.DIM;
	public static final int LEAF = Main.DIM;
	public static final int MAX_POINTS_IN_CELL = 4;

	//X,Y,Z, or LEAF
	public final int splitType;
	//pointers points if this is a leaf point
	public Point [] pointList;
	//split value if its not a leaf cell
	public final double splitValue;
	//else if its not a leaf point, we need children
	public volatile KdCell leftChild;
	public volatile KdCell rightChild;
	
	/**
	 * Create a new empty KDTree
	 */
	protected KdCell() {
		splitType = LEAF;
		splitValue = Float.MAX_VALUE;
		pointList = new Point[MAX_POINTS_IN_CELL];
		leftChild = null;
		rightChild = null;
	}

	//special constructor used internally

	protected KdCell(int inSplitType, double inSplitValue) {	  
		//we don't set the bounding box as we assume it will be set next
		splitType = inSplitType;
		splitValue = inSplitValue;
		pointList = (inSplitType == LEAF) ? new Point[MAX_POINTS_IN_CELL] : new Point[1];
	}

	protected static KdCell createNewBlankCell(int inSplitType, double inSplitValue) {
		return new KdCell(inSplitType, inSplitValue);
	}

	/**
	 * Sets the contents of this cell to the specified list and size.  Then sudivides the cell as
	 * necessary to build an appropriate hierarchy.  Val is a temporary array of doubles that
	 * we can pass in to reduce the allocation of additional temporary space.
	 */
	protected static KdCell subdivide(final Point list[], int offset, int size) {
		if (size <= MAX_POINTS_IN_CELL) {
			KdCell cell = createNewBlankCell(LEAF, Float.MAX_VALUE);
			System.arraycopy(list, offset, cell.pointList, 0, size);
			return cell;
		}
		
		//compute bounding box of points
		double [] min = new double[DIM];
		double [] max = new double[DIM];
		for (int i = 0; i < DIM; i++) {
			min[i] = Float.MAX_VALUE;
			max[i] = Float.MIN_VALUE;
		}
		for (int i = offset; i < size + offset; i++) {    	
			for (int j = 0; j < DIM; j++) {
				min[j] = Math.min(list[i].get(j), min[j]);
				max[j] = Math.max(list[i].get(j), max[j]);
			}
		}
		
		//choose split plane
		double [] diff = new double[DIM];
		for (int i = 0; i < DIM; i++) {
			diff[i] = max[i] - min[i];
		}
		int type = 0;
		double maxDiff = 0;
		for (int i = 0; i < DIM; i++) {
			if (diff[i] > maxDiff) {
				maxDiff = diff[i];
				type = i;
			}
		}
		
		Arrays.sort(list, offset, offset + size, new Point.PointComparable(type));
		int leftCount = size / 2;
		Point midPoint = list[offset + leftCount];
		
//		for (int k = offset; k < offset + size; k++) {
//			Point p = list[k];
//			System.out.printf("[%d] %.4f %.4f\n", k, p.coord[0], p.coord[1]);	
//		}
//		System.out.printf("splitval:%.4f type:%d leftCount:%d\n\n", midPoint.coord[type], type, leftCount);
		KdCell cell = createNewBlankCell(type, midPoint.coord[type]);
		cell.pointList[0] = list[offset + leftCount];
		cell.leftChild = subdivide(list, offset, leftCount);
		cell.rightChild = subdivide(list, offset + leftCount + 1, size - leftCount - 1);
		return cell;
	}

	public static KdCell createTree(Point inPoints[]) {
		KdCell root = subdivide(inPoints, 0, inPoints.length);
		return root;
	}

	public void findNearestNeighbor(Point point) {
		recurseNearestNeighbor(point, 0);
	}

	public void recurseNearestNeighbor(Point point, double splitDist) {
		double nextSplitDist = 0;
		point.nodesTraversed++;
		if (splitDist != 0 && point.kthDist() < splitDist * splitDist) {
			return;
		}
		if (splitType == KdCell.LEAF) {
			for (Point train : pointList) {
				if (train != null) {
					point.updateClosest(train);
				}
			}
		} else {
			point.updateClosest(pointList[0]);
			double split;
			
			if (splitType < DIM) {
				split = point.coord[splitType];
				nextSplitDist = split - splitValue;
			} else {
				throw new RuntimeException();
			}
			if (split <= splitValue) {
				leftChild.recurseNearestNeighbor(point, 0);
				rightChild.recurseNearestNeighbor(point, nextSplitDist);
			} else {
				rightChild.recurseNearestNeighbor(point, 0);
				leftChild.recurseNearestNeighbor(point, nextSplitDist);
			}
		}
	}    
}
