/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)
https://sites.google.com/site/treesplicer/

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.

These benchmarks are adapted from the Lonestar benchmark suite:
http://iss.ices.utexas.edu/?p=projects/galois/lonestar
 */

package knearestneighborsplitplane;

import java.util.Comparator;

public class Point {
	
	public static class PointComparable implements Comparator<Point> {
		int type;
		
		public PointComparable(int type) {
			this.type = type;
		}

		@Override
		public int compare(Point p0, Point p1) {
			return p0.coord[type] < p1.coord[type] ? -1 : (p0.coord[type] == p1.coord[type] ? 0 : 1); 
		}	
	}
	
	private final static int K = Main.K;
	private final static int maxLabels = Main.maxLabels;

	double [] coord;
	// track the Kth closets point for verification
	Point naiveKthClosest;
	Point [] closests;
	double [] dists;
	int kthIdx;
	int label;
	
	int nodesTraversed = 0;
	
	public Point(int label, double [] coord) {
		assert label < maxLabels;
		this.label = label;
		this.coord = coord;
		closests = new Point[K];
		dists = new double[K];
		for (int i = 0; i < K; i++) {
			dists[i] = Double.MAX_VALUE;
		}
	}

	public double get(int i) {
		return coord[i];
	}

	public double [] getCoords() {
		return coord;
	}  

	public double dist(Point dest) {
		double sum = 0;
		double dist;
		for (int i = 0; i < KdCell.DIM; i++) {
			dist = coord[i] - dest.coord[i];
			sum += dist * dist;
		}
		return sum;
	}

	public double kthDist() {
		return dists[kthIdx];
	}

	public void updateClosest(Point p) {
		if (p == this) {
			return;
		}
		double newDist = dist(p);
		if (newDist < dists[kthIdx]) {
			closests[kthIdx] = p;
			dists[kthIdx] = newDist;
			for (int i = 0; i < K; i++) {
				if (dists[i] > dists[kthIdx]) {
					kthIdx = i;
				}
			}
		}
	}
	
	public void markNaiveResult() {
		naiveKthClosest = closests[kthIdx];
		for (int i = 0; i < K; i++) {
			closests[i] = null;
			dists[i] = Double.MAX_VALUE;
		}
		kthIdx = 0;
	}
	
	public void verifyNaiveResult() {
		assert naiveKthClosest == closests[kthIdx];
	}
	
	public int majorityLabel() {
		int [] cnts = new int[maxLabels];
		for (int i = 0; i < K; i++) {
			int label = closests[i].label; 
			cnts[label]++;
		}
		int label = 0;
		for (int i = 0; i < maxLabels; i++) {
			if (cnts[i] > cnts[label]) {
				label = i;
			}
		}
		return label;
	}
	
	public String toString() {
		String ret = "";
		for (int i = 0; i < KdCell.DIM; i++) {
			ret += (coord[i] + "\t");
		}
		return ret;
	}
}
