/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)
https://sites.google.com/site/treesplicer/

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.

These benchmarks are adapted from the Lonestar benchmark suite:
http://iss.ices.utexas.edu/?p=projects/galois/lonestar
 */

package nearestneighbor;

import java.util.Arrays;

public class KdCell {

	public final static int DIM = Main.DIM;
	public final static int LEAF = Main.DIM;
	public static final int MAX_POINTS_IN_CELL = 4;

	//bounding box of points contained in this cell (and all descendents)
	protected double [] min;
	protected double [] max;

	//X,Y,Z, or LEAF
	public final int splitType;
	//pointers points if this is a leaf point
	public Point [] pointList;
	//split value if its not a leaf cell
	public final double splitValue;
	//else if its not a leaf point, we need children
	public volatile KdCell leftChild;
	public volatile KdCell rightChild;

	/**
	 * Create a new empty KDTree
	 */
	protected KdCell() {
		min = new double[DIM];
		max = new double[DIM];
		for (int i = 0; i < DIM; i++) {
			min[i] = Float.MAX_VALUE;
			max[i] = Float.MIN_VALUE;
		}
		splitType = LEAF;
		splitValue = Float.MAX_VALUE;
		pointList = new Point[MAX_POINTS_IN_CELL];
		leftChild = null;
		rightChild = null;
	}

	//special constructor used internally

	protected KdCell(int inSplitType, double inSplitValue) {
		min = new double[DIM];
		max = new double[DIM];	  
		//we don't set the bounding box as we assume it will be set next
		splitType = inSplitType;
		splitValue = inSplitValue;
		pointList = (inSplitType == LEAF) ? new Point[MAX_POINTS_IN_CELL] : null;
	}

	/**
	 * We provide this factory method so that KDCell can be subclassed.  Returns a new
	 * uninitialized cell (also tried to reuse any preallocated array for holding children)
	 * Used during cell subdivision.
	 */
	protected KdCell createNewBlankCell(int inSplitType, double inSplitValue) {
		return new KdCell(inSplitType, inSplitValue);
	}

	/**
	 * Computes this cells bounding box to just contain the specified points
	 */
	private void computeBoundingBoxFromPoints(Point list[], int size) {
		double [] minNew = new double[DIM];
		double [] maxNew = new double[DIM];	  
		for (int i = 0; i < DIM; i++) {
			minNew[i] = Float.MAX_VALUE;
			maxNew[i] = -Float.MAX_VALUE;
		}
		for (int i = 0; i < size; i++) {
			double [] coord = new double[DIM];
			for (int j = 0; j < DIM; j++) {
				coord[j] = list[i].get(j);
				minNew[j] = Math.min(coord[j], minNew[j]);
				maxNew[j] = Math.max(coord[j], maxNew[j]);
			}
		}
		for (int i = 0; i < DIM; i++) {
			min[i] = minNew[i];
			max[i] = maxNew[i];
		}
	}

	/**
	 * Return the appropriate splitting component (x,y, or z) which is relevant for this point
	 */
	private static double findSplitComponent(Point cluster, int splitType) {
		if (splitType < DIM) {
			return cluster.get(splitType);
		} else {
			throw new RuntimeException("badness");
		}

	}

	/**
	 * Given a list of points, and a split plane defined by the splitType and splitValue,
	 * partition the list into points below (<=) and above (>) the plane.  Returns the number of points
	 * which fell below the plane.
	 */
	private static int splitList(Point list[], int startIndex, int size, double splitValue, int splitType) {
		int lo = startIndex;
		int hi = startIndex + size - 1;
		//split into a low group that contains all points <= the split value and
		//a high group with all the points > the split value
		//note: after splitting, (lo - startIndex) will be the size of the low group
		while (lo <= hi) {
			while (lo <= hi && splitValue >= findSplitComponent(list[lo], splitType)) {
				lo++;
			}
			while (lo <= hi && splitValue < findSplitComponent(list[hi], splitType)) {
				hi--;
			}
			if (lo < hi) {
				int index1 = lo++;
				int index2 = hi--;
				Point temp = list[index1];
				list[index1] = list[index2];
				list[index2] = temp;
			}
		}
		return lo - startIndex;
	}

	/**
	 * Sets the contents of this cell to the specified list and size.  Then sudivides the cell as
	 * necessary to build an appropriate hierarchy.  Val is a temporary array of doubles that
	 * we can pass in to reduce the allocation of additional temporary space.
	 */
	protected static KdCell subdivide(final Point list[], int offset, int size, double doubleArr[], KdCell factory) {
		if (size <= MAX_POINTS_IN_CELL) {
			KdCell cell = factory.createNewBlankCell(LEAF, Float.MAX_VALUE);
			System.arraycopy(list, offset, cell.pointList, 0, size);
			cell.computeBoundingBoxFromPoints(cell.pointList, size);
			return cell;
		}
		//otherwise its an interior point and we need to choose a split plane
		if (doubleArr == null) {
			doubleArr = new double[size];
		}
		//compute bounding box of points
		double [] min = new double[DIM];
		double [] max = new double[DIM];
		for (int i = 0; i < DIM; i++) {
			min[i] = Float.MAX_VALUE;
			max[i] = Float.MIN_VALUE;
		}
		for (int i = offset; i < size + offset; i++) {    	
			for (int j = 0; j < DIM; j++) {
				min[j] = Math.min(list[i].get(j), min[j]);
				max[j] = Math.max(list[i].get(j), max[j]);
			}
		}
		//choose split plane

		double [] s = new double[DIM];
		int [] order = new int[DIM];
		boolean [] chosen = new boolean[DIM];
		for (int i = 0; i < DIM; i++) {
			s[i] = max[i] - min[i];
			chosen[i] = false;
		}

		for (int i = 0; i < DIM; i++) {
			double tmp = Float.MIN_VALUE;
			int idx = 0;
			for (int j = 0; j < DIM; j++) {
				if (s[j] > tmp && !chosen[j]) {
					idx = j;
					tmp = s[j];
				}
			}
			order[i] = idx;
			chosen[idx] = true;
		}

		int type = 0;
		double value = 0;
		for (int i = 0; i < DIM; i++) {
			type = order[i];
			value = computeSplitValue(list, offset, size, type, doubleArr);
			if (value != Float.MAX_VALUE) {
				break;
			}
		}
		if (value == Float.MAX_VALUE) {
			throw new RuntimeException("badness splittype:" + type + " value:" + value + " size:" + size);
		}
		int leftCount = splitList(list, offset, size, value, type);
		if (leftCount <= 1 || leftCount >= size - 1) {
			throw new RuntimeException("badness splittype:" + type + " value:" + value + " size:" + size + " leftCount:" + leftCount);
		}
		KdCell cell = factory.createNewBlankCell(type, value);
		for (int i = 0; i < DIM; i++) {
			cell.min[i] = min[i];
			cell.max[i] = max[i];
		}
		cell.leftChild = subdivide(list, offset, leftCount, doubleArr, factory);
		cell.rightChild = subdivide(list, offset + leftCount, size - leftCount, doubleArr, factory);
		return cell;
	}

	private static double computeSplitValue(Point list[], int offset, int size, int splitType, double doubleArr[]) {
		for (int i = 0; i < size; i++) {
			doubleArr[i] = findSplitComponent(list[offset + i], splitType);
		}
		return findMedianGapSplit(doubleArr, size);
	}

	/**
	 * Given an array of doubles, sorts the list, finds the largest gap in values
	 * near the median, and returns a value in the middle of that gap
	 */
	private static double findMedianGapSplit(double val[], int size) {
		//this is not very efficient at the moment, there are faster median finding algorithms
		Arrays.sort(val, 0, size);
		int start = ((size - 1) >> 1) - ((size + 7) >> 3);
		int end = (size >> 1) + ((size + 7) >> 3);
		if (start == end) {
			//should never happen
			throw new RuntimeException();
		}
		double largestGap = 0;
		double splitValue = 0;
		double nextValue = val[start];
		for (int i = start; i < end; i++) {
			double curValue = nextValue; //ie val[i]
			nextValue = val[i + 1];
			if ((nextValue - curValue) > largestGap) {
				largestGap = nextValue - curValue;
				splitValue = 0.5f * (curValue + nextValue);
				if (splitValue == nextValue) {
					splitValue = curValue;
				} //if not between then choose smaller value
			}
		}
		if (largestGap <= 0) {
			//indicate that the attempt to find a good split value failed
			return Float.MAX_VALUE;
		}
		return splitValue;
	}

	public static KdCell createTree(Point inPoints[]) {
		KdCell root = subdivide(inPoints, 0, inPoints.length, null, new KdCell());
		return root;
	}

	public void findNearestNeighbor(Point point) {
		recurseNearestNeighbor(point);
	}

	public void recurseNearestNeighbor(Point point) {
		point.nodesTraversed++;
		if (!canCorrelate(point, point.closestDist)) {
			return;
		}
		if (splitType == KdCell.LEAF) {
			for (Point train : pointList) {
				if (train != null) {
					point.updateClosest(train);
				}
			}
		} else {
			double split;
			if (splitType < DIM) {
				split = point.coord[splitType];
			} else {
				throw new RuntimeException();
			}
			if (split <= splitValue) {
				leftChild.recurseNearestNeighbor(point);
				rightChild.recurseNearestNeighbor(point);
			} else {
				rightChild.recurseNearestNeighbor(point);
				leftChild.recurseNearestNeighbor(point);
			}  			
		}
	}  

	private boolean canCorrelate(Point point, double rad) {
		boolean ret = false;
		double sum = 0;
		double boxsum = 0;
		for (int i = 0 ; i < DIM; i++) {
			double center = (max[i] + min[i]) / 2;
			double boxdist = (max[i] - min[i]) / 2;
			double dist = point.coord[i] - center;
			sum += dist * dist;
			boxsum += boxdist * boxdist;
		}
		if (Math.sqrt(sum) - Math.sqrt(boxsum) < Math.sqrt(rad)) {
			ret = true;
		}
		return ret;
	}  
}
