/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)
https://sites.google.com/site/treesplicer/

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.

These benchmarks are adapted from the Lonestar benchmark suite:
http://iss.ices.utexas.edu/?p=projects/galois/lonestar
 */

package nearestneighbor;


import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Random;
import java.util.Scanner;
import java.util.concurrent.ExecutionException;
import java.util.zip.GZIPInputStream;

import util.Launcher;

public class Main {

	// verifies nearest neighbors are consistent with naive O(N^2) algorithm
	static final boolean verify = false;

	static final boolean readTestDataFirst = false;
	static final boolean randomizeTestData = false;

	static int DIM;

	public static void main(String[] args) {
		if (args.length != 5) {
			System.err.println("arguments: sort dimensionality numTrainPoints numTestPoints [filename]");
			System.err.println("ex) false 3 10000 10000 random");
			System.err.println("ex) false 7 10000 10000 input/nearestneighbor/covtype.7d.gz");
			System.exit(-1);
		}
		boolean sort = Boolean.parseBoolean(args[0]);
		DIM = Integer.parseInt(args[1]);
		//KdCell.setDimension(Integer.parseInt(args[1]));
		int numTrainPoints = Integer.parseInt(args[2]);
		int numTestPoints = Integer.parseInt(args[3]);
		String filename = args[4];
		Point [] trainData = new Point[numTrainPoints];
		Point [] testData = new Point[numTestPoints];
		if (filename.equals("random")) {
			Random rand = new Random(0);
			genData(trainData, rand);
			genData(testData, rand);			
		} else {
			readInput(filename, trainData, testData);
		}
		float rate = process(trainData, testData, sort, filename);
		recordCorrectRate(args, rate);
	}

	static void recordCorrectRate(String [] args, float rate) {
		String argstr = "";
		for (int i = 0; i < args.length; i++) {
			argstr += args[i] + " ";
		} 
		PrintStream csv;
		try {
			csv = new PrintStream(new FileOutputStream("nncorrect.csv", true));
			csv.printf("%s, %.4f\n", argstr, rate);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	static void readInput(String filename, Point [] trainData, Point [] testData) {
		Scanner in = null;
		try {
			if (filename.endsWith(".gz")) {
				in = new Scanner(new GZIPInputStream(new FileInputStream(filename)));
			} else {
				in = new Scanner(new BufferedReader(new FileReader(filename)));
			}
			in.useLocale(Locale.US);
		} catch (FileNotFoundException e) {
			System.err.println(e);
			System.exit(-1);
		} catch (IOException e) {
			System.err.println(e);
			System.exit(-1);
		}
		if (readTestDataFirst) {
			readData(testData, in);
			readData(trainData, in);
		} else {
			readData(trainData, in);
			readData(testData, in);
		}
		if (randomizeTestData) {
			Random rand = new Random(0);
			for (int i = 0; i < testData.length; i++) {
				int r = rand.nextInt(testData.length);
				Point p = testData[i];
				testData[i] = testData[r];
				testData[r] = p;
			}
		}
	}

	static void readData(Point [] data, Scanner in) {
		int idx = 0;
		do {
			int label = in.nextInt();
			double [] coord = new double[KdCell.DIM];
			for (int i = 0; i < KdCell.DIM; i++) {
				coord[i] = in.nextDouble();
			}
			data[idx++] = new Point(label, coord);
		} while (idx < data.length);	
	}

	static void genData(Point [] points, Random rand) {
		for(int i = 0; i < points.length; i++) {
			double [] coord = new double[KdCell.DIM];
			for(int j = 0; j < KdCell.DIM; j++) {
				coord[j] = rand.nextDouble();
			}
			points[i] = new Point(0, coord);
		}
	}

	static void naiveNearestNeighbor(final Point[] trainData, final Point[] testData) {
		for (Point test : testData) {
			for (Point train : trainData) {
				test.updateClosest(train);
			}
			test.naiveClosest = test.closest;
			test.closest = null;
			test.closestDist = Double.MAX_VALUE;
		}
	}

	static void labelStats(Point [] points, int labels) {
		int [] cnts = new int[labels];
		for (Point p : points) {
			cnts[p.label]++;
		}
		for (int i = 0; i < labels; i++) {
			System.out.printf("%d: %d\n", i, cnts[i]);
		}
	}

	static float process(final Point[] trainData, final Point[] testData, boolean sort, String filename) {
		if (verify) {
			naiveNearestNeighbor(trainData, testData);
		}

		KdCell root = KdCell.createTree(trainData);

		Launcher.startTiming();
		if (sort) {
			// building kdtree sorts the points
			KdCell.createTree(testData);
		}

		try {
			ParallelFor(root, testData);
		} catch (ExecutionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		Launcher.stopTiming();

		int correct = 0;
		for (Point p : testData) {
			if (p.closest.label == p.label) {
				correct++;
			}
		}
		float correctRate = ((float) correct) / testData.length;
		System.out.printf("correct: %.4f\n", correctRate);

		// check expected values
		long totalNodesTraversed = 0;
		for (int i = 0; i < testData.length; i++) {
			totalNodesTraversed += testData[i].nodesTraversed;
		}
		System.out.printf("totalNodesTraversed: %d\n", totalNodesTraversed);

		// check expected values
		if (trainData.length == 10000 && testData.length == 10000 && DIM == 7 && filename.equals("random")) {
			if (totalNodesTraversed == 6931736) {
				System.out.println("expected values passed");
			} else {
				System.out.println("expected values failed");
				assert false;
			}
		}

		if (verify) {
			for (Point test : testData) {
				assert test.naiveClosest == test.closest;
			}
			System.out.println("verify successful");
		}	

		return correctRate;
	}

	// a substitute for parallel for annotations on for statements, which are not supported in Java 6
	public @interface ParallelFor {}
	@ParallelFor
	public static void ParallelFor(final KdCell root, final Point [] testData) throws ExecutionException {
		for (int i = 0; i < testData.length; i++) {
			Point test = testData[i];
			root.findNearestNeighbor(test);
		}
	}
}
