/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)
https://sites.google.com/site/treesplicer/

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.

These benchmarks are adapted from the Lonestar benchmark suite:
http://iss.ices.utexas.edu/?p=projects/galois/lonestar
 */

package pointcorrelation;


import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Random;
import java.util.Scanner;
import java.util.concurrent.ExecutionException;
import java.util.zip.GZIPInputStream;

import util.Launcher;

public class Main {

	// verifies nearest neighbors are consistent with naive O(N^2) algorithm
	static final boolean verify = false;
	// checks total number of nodes traversed are consistent with expected value
	static final boolean checkNumNodesTraversed = true;

	static int DIM;

	public static void main(String[] args) {
		if (args.length != 5) {
			System.err.println("arguments: sort dimensionality radius numPoints [filename]");
			System.err.println("ex) false 3 0.05 10000 random");
			System.err.println("ex) false 7 0.001 10000 input/nearestneighbor/covtype.7d.gz");
			System.exit(-1);
		}
		boolean sort = Boolean.parseBoolean(args[0]);
		DIM = Integer.parseInt(args[1]);
		double radius = Float.parseFloat(args[2]);
		int numPoints = Integer.parseInt(args[3]);
		String filename = args[4];
		Point [] points = new Point[numPoints];
		if (filename.equals("random")) {
			Random rand = new Random(0);
			genData(points, rand);
		} else {
			readInput(filename, points);
		}
		correlation(points, radius, sort, filename);
	}

	static void readInput(String filename, Point [] points) {
		Scanner in = null;
		try {
			if (filename.endsWith(".gz")) {
				in = new Scanner(new GZIPInputStream(new FileInputStream(filename)));
			} else {
				in = new Scanner(new BufferedReader(new FileReader(filename)));
			}
			in.useLocale(Locale.US);
		} catch (FileNotFoundException e) {
			System.err.println(e);
			System.exit(-1);
		} catch (IOException e) {
			System.err.println(e);
			System.exit(-1);
		}
		readData(points, in);
	}

	static void readData(Point [] data, Scanner in) {
		int idx = 0;
		do {
			int label = in.nextInt();
			double [] coord = new double[KdCell.DIM];
			for (int i = 0; i < KdCell.DIM; i++) {
				coord[i] = in.nextDouble();
			}
			data[idx++] = new Point(label, coord);
		} while (idx < data.length);	
	}

	static void genData(Point [] points, Random rand) {
		for(int i = 0; i < points.length; i++) {
			double [] coord = new double[KdCell.DIM];
			for(int j = 0; j < KdCell.DIM; j++) {
				coord[j] = rand.nextDouble();
			}
			points[i] = new Point(0, coord);
		}
	}

	// naive N^2 correlation search, used for verification only
	static void naiveCorrelation(final Point[] points, double rad) {
		for (Point p : points) {
			for (Point p2 : points) {
				if (p.inRadii(p2, rad)) {
					p.naiveCorr++;
				}
			}
		}
	}

	private static void correlation(final Point[] points, double rad, boolean sort, String filename) {
		if (verify) {
			naiveCorrelation(points, rad);
		}

		KdCell root; 
		if (!sort) {
			Point [] buildPoints = new Point[points.length];
			for (int i = 0; i < points.length; i++) {
				buildPoints[i] = points[i];
			}
			root = KdCell.createTree(buildPoints);
		} else {
			// building the kd-tree sorts the points
			root = KdCell.createTree(points);
		}

		Launcher.startTiming();
		try {
			ParallelFor(root, points, rad);
		} catch (ExecutionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		Launcher.stopTiming();

		long sum = 0;
		for (Point p : points) {
			sum += p.corr;
		}
		System.out.printf("correlation: %.2f\n", ((float) sum) / points.length);



		long totalNodesTraversed = 0;
		for (int i = 0; i < points.length; i++) {
			totalNodesTraversed += points[i].nodesTraversed;
		}
		System.out.printf("totalNodesTraversed: %d\n", totalNodesTraversed);


		// check expected values
		if (points.length == 10000 && rad == 0.1f && DIM == 3 && filename.equals("random")) {
			if (totalNodesTraversed == 1746244) {
				System.out.println("expected values passed");
			} else {
				System.out.println("expected values failed");
				assert false;
			}
		}

		if (verify) {
			for (int i = 0; i < points.length; i++) {
				Point ln = points[i];
				assert ln.naiveCorr == ln.corr;
			}
		}
	}

	// a substitute for parallel for annotations on for statements, which are not supported in Java 6
	public @interface ParallelFor {}
	@ParallelFor
	public static void ParallelFor(final KdCell root, final Point [] points, final double rad) throws ExecutionException {

		for (int i = 0; i < points.length; i++) {
			Point ln = points[i];
			root.findCorrelation(ln, rad);
		}	
	}
}
