/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.
 */

package transform;

import java.util.ArrayList;

import AST.ASTNode;
import AST.Access;
import AST.Annotation;
import AST.ArrayAccess;
import AST.ArrayDecl;
import AST.Block;
import AST.BodyDecl;
import AST.CastExpr;
import AST.ClassDecl;
import AST.CompilationUnit;
import AST.Dot;
import AST.Expr;
import AST.FieldDeclaration;
import AST.ForStmt;
import AST.IfStmt;
import AST.MethodAccess;
import AST.MethodDecl;
import AST.Modifier;
import AST.Modifiers;
import AST.ParExpr;
import AST.PrimitiveTypeAccess;
import AST.SimpleSet;
import AST.Stmt;
import AST.TypeAccess;
import AST.TypeDecl;
import AST.VarAccess;

public class RecursiveStructureAnalysis {
	/**
	 * first gather access points and recursive fields
	 * @param cu
	 */
	public static void processPre(CompilationUnit cu) {
		cu.gatherAccessPoints();
		if (cu.getNumTypeDecl() != 1) {
			System.out.println(cu);
			for (TypeDecl tdec : cu.getTypeDeclList()) {
				System.out.println(tdec + "\n");
			}
			assert false;
		}
		TypeDecl tdec = cu.getTypeDecl(0);
		if (tdec instanceof ClassDecl) {
			ClassDecl cdec = (ClassDecl) cu.getTypeDecl(0);
			cdec.gatherRecursiveFields();
		}

		String relName = cu.relativeName();
		String [] split = relName.split("/");
		String packageName = "";
		for (int i = 1; i < split.length - 1; i++) {
			packageName += split[i];
			if (i != split.length - 2) {
				packageName += ".";  
			}
		}
		if (!packageName.equals("")) {
			Globals.packageNames.add(packageName);
		}
	}

	/**
	 * next look for recursive structures
	 * @param cu
	 */
	public static void process(CompilationUnit cu) {
		TypeDecl tdec = cu.getTypeDecl(0);
		if (tdec instanceof ClassDecl) {
			System.out.println("processing " + cu.pathName());
			// for now, just find one recursive structure, look for only if not found yet
			if (Globals.mainPath.isEmpty()) {
				findRecursiveStructure((ClassDecl) cu.getTypeDecl(0));
				if (!Globals.mainPath.isEmpty()) {
					setBlockPaths(cu);
				}
			}
		}
	}
	
	private static void findRecursiveStructure(ClassDecl classDec) {	
		for (BodyDecl bdec : classDec.getBodyDeclList()) {
			if (bdec instanceof MethodDecl) {
				MethodDecl mdec = (MethodDecl) bdec;
				if (mdec.isRecursive()) {
					if (findRecursiveUse(mdec)) {
						processRecursiveStructure(mdec);
						Globals.isImplicitRecursiveStructure = false;
						break;
					} else if (mdec.getNumParameter() > 0) {
						if (classDec.isRecursive()){
							if (findRecursiveImplicitUse(mdec, classDec)) {
								processRecursiveStructure(mdec);
								Globals.isImplicitRecursiveStructure = true;
								break;
							}
						}
					}
				}
			}
		}
	}

	private static boolean findRecursiveImplicitUse(MethodDecl mdec, ClassDecl cdec) {
		boolean ret = false;
		RecursiveField.clearFields();
		for (MethodAccess macc : mdec.accessPoints) {
			if (Utils.getEnclosingMethodDecl(macc) == mdec) {
				ASTNode node = macc.getParent();
				if (node instanceof Dot) {
					Dot dot = (Dot) node;
					Expr left = dot.getLeft();
					if (left instanceof ParExpr) {
						ParExpr pexpr = (ParExpr) left;
						left = pexpr.getExpr();
					}
					if (left instanceof CastExpr) {
						CastExpr cexpr = (CastExpr) left;
						left = cexpr.getExpr();
					}
					if (left instanceof VarAccess) {
						VarAccess lvacc = (VarAccess) left;
						PointsToAnalysis pointsTo = new PointsToAnalysis(mdec);
						if (pointsTo.pointsToMany(lvacc)) {
							return false;
						}
						TypeDecl ltdec = lvacc.type();
						if (ltdec == cdec) {
							for (FieldDeclaration fdec : cdec.recursiveFields) {
								if (fdec.getID().equals(lvacc.getID())) {
									ret = true;
									RecursiveField.addRecursiveField(fdec, cdec);
								}
							}
						} else {
							while (cdec.hasSuperclass()) {
								cdec = cdec.superclass();
								if (ltdec == cdec) {
									for (FieldDeclaration fdec : cdec.recursiveFields) {
										if (fdec.getID().equals(lvacc.getID())) {
											ret = true;
											RecursiveField.addRecursiveField(fdec, cdec);
										}
									}								
								}
							}
						}
					}
				}
			}
		}
		return ret;
	}

	private static boolean findRecursiveUse(MethodDecl mdec) {
		boolean ret = false;
		RecursiveField.clearFields();
		if (mdec.getNumParameter() <= 1) {
			return ret;
		}
		if (mdec.findRecursiveVarDef()) {
			return ret;
		}

		for (MethodAccess macc : mdec.accessPoints) {
			if (Utils.getEnclosingMethodDecl(macc) == mdec) {
				// this is recursion point
				for (Expr arg : macc.getArgs()) {
					arg = Utils.getMainExpr(arg);
					/*
					 * if arg is local variable, get value it was initialized with
					 * cf) Node child = nn.children[treeIndex];
					 * to be general, need reaching def analysis
					 */
					if (arg instanceof VarAccess) {
						VarAccess vacc = (VarAccess) arg;
						PointsToAnalysis pointsTo = new PointsToAnalysis(mdec);
						Expr ptexpr = pointsTo.getFinalPointsTo(vacc);
						if (ptexpr == null) {
							return false;
						}
						arg = ptexpr;
					}

					if (arg instanceof Dot) {
						Dot dot = (Dot) arg;
						Expr left = dot.getLeft();
						TypeDecl ltdec = null;
						/*
						 * if left arg was casted, use casted type
						 */
						if (left instanceof ParExpr) {
							left = ((ParExpr) left).getExpr();
							if (left instanceof CastExpr) {
								CastExpr cexpr = (CastExpr) left;
								left = cexpr.getExpr();
								ltdec = cexpr.type();
							}
						}
						Expr right = dot.getRight();
						// get var part for arrays
						if (right instanceof Dot) {
							dot = (Dot) right;
							if (dot.getRight() instanceof ArrayAccess) {
								right = dot.getLeft();
							}
						}
						if (left instanceof VarAccess && right instanceof VarAccess) {
							VarAccess lvacc = (VarAccess) left;
							VarAccess rvacc = (VarAccess) right;
							/*
							 * if left arg was not casted, use original type
							 */
							if (ltdec == null) {
								ltdec = lvacc.type();
							}
							if (ltdec instanceof ArrayDecl) {
								ArrayDecl adec = (ArrayDecl) ltdec;
								ltdec = (TypeDecl) adec.getParent();
							}
							if (ltdec instanceof ClassDecl) {
								ClassDecl cdec = (ClassDecl) ltdec;
								for (FieldDeclaration fdec : cdec.recursiveFields) {
									if (fdec.getID().equals(rvacc.getID())) {
										ret = true;
										RecursiveField.addRecursiveField(fdec, cdec);
									}
								}
								while (cdec.hasSuperclass()) {
									cdec = cdec.superclass();
									for (FieldDeclaration fdec : cdec.recursiveFields) {
										if (fdec.getID().equals(rvacc.getID())) {
											ret = true;
											RecursiveField.addRecursiveField(fdec, cdec);
										}
									}
								}
							}
						}
					}
				}
			}
		}
		return ret;
	}

	private static void processRecursiveStructure(MethodDecl mdec) {
		System.out.printf("\t candidate recursive method: %s\n", mdec.getID());
		boolean found = false;
		boolean recursiveEntryFound = false;
		TypeAccess retTacc = (TypeAccess) mdec.getTypeAccess();
		if (!retTacc.getID().equals("void")) {
			System.out.printf("\t\t skipped because return type is not void [%s]\n", retTacc.getID());
			return;
		}
		if (!isPseudoTailRecursive(mdec)) {
			System.out.printf("\t\t skipped because not pseudo-tail recursive\n");
			return;			
		}
		for (MethodAccess macc : mdec.accessPoints) {
			if (Utils.getEnclosingMethodDecl(macc) != mdec) {
				// currently assuming one access point for the recursive entry
				assert !found;
				found = true;
				if (findEnclosingForLoop(macc)) {
					Globals.mainPath.add(macc);
					recursiveEntryFound = true;
				}
			}
		}
		if (!recursiveEntryFound) {
			System.out.printf("\t\t skipped because enclosing parallel for loop not found\n");
			Globals.mainPath.clear();
		} else {
			System.out.printf("\t\t selected for transformation\n");
			// reverse mainPath to match usage throughout rest of program
			ArrayList<MethodAccess> temp = new ArrayList<MethodAccess>();
			for (MethodAccess macc : Globals.mainPath) {
				temp.add(macc);
			}
			Globals.mainPath.clear();
			for (int i = temp.size() - 1; i >= 0; i--) {
				Globals.mainPath.add(temp.get(i));
			}
		}
	}
	
	private static boolean findEnclosingForLoop(MethodAccess macc) {
		ASTNode node = macc;
		while (node != null) {
			node = node.getParent();
			if (node instanceof ForStmt) {
				MethodDecl encMdec = Utils.getEnclosingMethodDecl(node);
				// transform only if enclosing method has @Parallel annotation
				boolean parallelAnnotation = false;
				for (Modifier mod : encMdec.getModifiers().getModifierList()) {
					if (mod instanceof Annotation) {
						Annotation annot = (Annotation) mod;
						Access acc = annot.getAccess();
						if (acc instanceof TypeAccess) {
							TypeAccess tacc = (TypeAccess) acc;
							if (tacc.getID().equals(Utils.annotationName_parallel)) {
								parallelAnnotation = true;
								break;
							}
						}
					}
				}
				if (parallelAnnotation) {
					Globals.outerFor = (ForStmt) node;
					return true;
				}
			} else if (node instanceof MethodDecl) {
				MethodDecl nextMdec = (MethodDecl) node;
				for (MethodAccess nextMacc : nextMdec.accessPoints) {
					if (findEnclosingForLoop(nextMacc)) {
						Globals.mainPath.add(nextMacc);
						return true;
					}
				}
			}
		}
		return false;
	}
	
	private static boolean isPseudoTailRecursive(MethodDecl mdec) {
		ArrayList<Stmt> recursionStmts = new ArrayList<Stmt>();
		for (MethodAccess macc : mdec.accessPoints) {
			if (Utils.getEnclosingMethodDecl(macc) != mdec) continue;
			recursionStmts.add(Utils.getEnclosingStmt(macc));
		}
		for (Stmt recursionStmt : recursionStmts) {
			ASTNode node = recursionStmt;
			while (node != null) {
				node = node.getParent();
				if (node instanceof Block) {
					Block block = (Block) node;
					boolean found = false;
					for (int i = 0; i < block.getNumStmt(); i++) {
						Stmt stmt = block.getStmt(i);
						if (stmt == recursionStmt) {
							found = true;
						} else if (found) {
							if (!recursionStmts.contains(stmt)) {
								System.out.printf(" [%s]\n followed by\n [%s]\n", recursionStmt, stmt);
								return false;
							}
						}
					}
					recursionStmt = block;
				} else if (node instanceof ForStmt) {
					System.out.printf("\t\t %s within for loop\n", recursionStmt);
					return false;
				} else if (node instanceof IfStmt) {
					recursionStmt = (Stmt) node;
				} else if (node instanceof MethodDecl) {
					break;
				}
			}
		}
		return true;
	}
	
	private static void setBlockPaths(CompilationUnit cu) {
		String relName = cu.relativeName();
		String [] split = relName.split("/");
		for (int i = 1; i < split.length - 1; i++) {
			Globals.blockPathName += split[i];
			Globals.blockPackageName += split[i];
			if (i != split.length - 2) {
				Globals.blockPathName += "/";
				Globals.blockPackageName += ".";  
			}
		}	
	}
}
